//------------------------------------------------------------------------
// zd_ok_la.v
//
// This code is the Zietlow Design implementation of the Opal Kelly "logic analyzer in a day" challenge.
//
// The design is based on the XEM-3005 module and uses the on-board SDRAM for data storage.
//	The SDRAM controller used is provided by Opal Kelly and is a basic page oriented controller.
//	Some of the code is borrowed from the Opal Kelly ramtest.v example.
//
//	The 16 bit data is sampled on every sample clock.
//	The sample clock can either be an external signal up to about 125 MHz or an internally generated
//	clock of selectable frequency. (100 MHz, 50 Mhz, 20 Mhz, 10 MHz ...)
//	The sampled data is synchronized to the SDRAM clock domain via a FIFO.
//	There it is compared to a trigger pattern which consists of two masks and to state words.
//	Data at t0 is masked by the first mask and the compared to the first state word while data at t1
//	is masked by the second mask and compared to the second state word. If there is a match of two
//	consecutive samples a trigger occurs. This scheme allows for any combination of 0, 1, X and transitions to
//	generate a trigger.
//
//	The logic analyzer has a few basic states it can be in.
//	halted		-	this state allows acquired data to be read by the host
//	pre-trigger	-	data is being sampled and stored, but trigger conditions are ignored
//	armed			-	the analyzer has stored the requested number of pre-trigger samples and now is actively 
//						waiting for a trigger. Acquisition continues in this state.
//	triggered	-	A trigger has occured and the logic analyzer is filling its post-trigger memory before
//						returning to the halted state.
//
// The trigger location is stored in a register so the host can align the acquired data appropriately.
//
//	To control the acquisition, the host writes a couple of registers holding the pre and post trigger sample sizes.
//	For simplicity they have a granularity of one SDRAM block (512 words).
//
// Host Interface registers:
// WireIn 0x00	clock settings and control bits
//		0 - 7	clock divider
//		8 - external clock select
//		9 - test pattern enable
//		10 - reset
// WireIn 0x01	pre-trigger count in SDRAM blocks
// WireIn 0x02	post-trigger count in SDRAM blocks
// WireIn 0x03	pattern_a
// WireIn 0x04	pattern_b
// WireIn 0x05	pattern_mask
// WireIn 0x06	starting address for SDRAM reads; bit 15 is used to reset the read FIFO
//
//	WireOut 0x20
//		0 - 8		trigger address lower 9 bits
//		12			halted
//		13			pre-trigger
//		14			armed
//		15			triggered
//	WireOut 0x21
//		0 - 14	trigger address page
//
//	TriggerIn 0x40
//		0 - set read address
//		1 - run
//		
// PipeOut 0xA0 - SDRAM read port
//
//------------------------------------------------------------------------

`timescale 1ns / 1ps

module zd_ok_la(
   input  wire [7:0]  hi_in,
   output wire [1:0]  hi_out,
   inout  wire [15:0] hi_inout,

   output wire        hi_muxsel,
	output wire        i2c_scl,
	output wire        i2c_sda,
	
	input  wire        clk1,	//	125 MHz SDRAM clock
	input	 wire			 gclk10,	//	external clock source
	input	 wire	[15:0] d,		//	logic analyzer inputs
	
	output wire			 sdram_cke,
	output wire        sdram_cs_n,
	output wire        sdram_we_n,
	output wire        sdram_cas_n,
	output wire        sdram_ras_n,
	output wire        sdram_ldqm,
	output wire        sdram_udqm,
	output wire [1:0]  sdram_ba,
	output wire [12:0] sdram_a,
	inout  wire [15:0] sdram_d,
   
	output wire [3:0]  led
);


// Host interface connections
wire        ti_clk;
wire [30:0] ok1;
wire [16:0] ok2;
wire [15:0] ep00wire;
wire [7:0]  clk_div = ep00wire[7:0];
wire			ext_sel = ep00wire[8];
reg			reset = 1'b0;
wire [15:0] pre_trigger_count;
wire [15:0] post_trigger_count;
wire [15:0] pattern_a;
wire [15:0] pattern_b;
wire [15:0] pattern_mask;
wire [15:0] read_address;
wire [15:0] ep40_triggers;
wire			set_read_address = ep40_triggers[0];
wire			run = ep40_triggers[1];
reg         halted = 1'b1;
reg         pre_trigger = 1'b0;
reg         armed = 1'b0;
reg         triggered = 1'b0;

wire        sdram_clk;

// SDRAM controller / negotiator connections
reg         cmd_pageread;
reg         cmd_pagewrite;
wire        cmd_ack;
wire        cmd_done;
reg  [14:0] rowaddr;

// SDRAM controller / FIFO connections.
wire        out_fifo_write;
wire [15:0] out_fifo_din;
wire        epA0_read;
wire [15:0] epA0_din;
wire [10:0] epA0fifo_status;

assign hi_muxsel = 1'b0;
assign i2c_scl = 1'bz;
assign i2c_sda = 1'bz;

assign sdram_cke = 1'b1;
assign sdram_ldqm = 1'b0;
assign sdram_udqm = 1'b0;

//reg	[25:0]	blinky;
//always @(negedge sdram_clk) begin
//	blinky	<=	blinky + 1;
//end
//assign led = ~{halted, pre_trigger, blinky[25], reset};

assign led = ~{halted, pre_trigger, armed, triggered};

//	sample cliock generation
wire	sample_clock;
wire	clk200;
sample_clock_gen sample_clock_gen_i (
    .clk200(clk200), 
    .ext_clk(gclk10), 
    .ext_sel(ext_sel), 
    .clk_div(clk_div), 
    .sample_clock(sample_clock)
);

//	test pattern generator
reg test_en = 1'b0;
always @(posedge sample_clock) begin
	test_en	<= ep00wire[9];
end 

wire	[15:0]	test_data;
test_pattern test_pattern_i (
    .sample_clock(sample_clock), 
    .test_en(test_en), 
    .test_data(test_data)
);

//	sample logic analyzer data inputs
(* IOB = "TRUE" *) reg	[15:0]	din = 16'd0;
reg	[15:0]	sample_data = 16'd0;
reg				in_fifo_wr_en = 1'b0;
always @(posedge sample_clock) begin
	din				<=	d;
	sample_data		<=	test_en ? test_data : din;
	in_fifo_wr_en	<=	!halted;
end

//-----------------------------------------------
//	input FIFO; data is loaded at sample clock rate and read at SDRAM clock rate
wire				in_fifo_read;
wire	[15:0]	in_fifo_dout;
wire	[10:0]	in_fifo_status;
fifo_2kx16 in_fifo (
	.rst				(reset),
	.wr_clk			(sample_clock),
	.rd_clk			(!sdram_clk),
	.din				(sample_data),
	.wr_en			(in_fifo_wr_en),
	.rd_en			(in_fifo_read),
	.dout				(in_fifo_dout),
	.full				(),
	.empty			(),
	.rd_data_count	(in_fifo_status),
	.wr_data_count	()
);

//	trigger pattern match logic
reg				trigger = 1'b0;
reg	[15:0]	comp_data = 16'd0;
reg				match_a = 1'b0;
reg				in_fifo_read_d = 1'b0;
reg				in_fifo_read_d2 = 1'b0;
reg	[15:0]	edge_pattern = 16'd0;
always @(negedge sdram_clk) begin
	trigger				<=	1'b0;
	in_fifo_read_d2	<= in_fifo_read_d;
	in_fifo_read_d		<= in_fifo_read;
	edge_pattern		<=	pattern_a ^ pattern_b;
	if(in_fifo_read_d) begin
		comp_data		<=	in_fifo_dout & pattern_mask;
	end
	if(in_fifo_read_d2) begin
		if(comp_data  == pattern_a) begin
			match_a		<=	1'b1;
		end
		else begin
			match_a		<=	1'b0;
		end
		if(((comp_data & edge_pattern) == (pattern_b & edge_pattern)) && (match_a == 1'b1)) begin
			trigger		<=	1'b1;
		end
	end
end

//	trigger address counter
//	keeps track of the trigger position within a page
reg	[8:0]	trigger_address_counter = 9'd0;
always @(negedge sdram_clk) begin
	if(cmd_ack) begin
		trigger_address_counter	<=	9'd0;
	end
	else if(in_fifo_read) begin
		trigger_address_counter	<=	trigger_address_counter + 9'd1;
	end
end

//-----------------------------------
//	acquisition state machine
reg	[14:0]	trigger_page = 15'd0;
reg	[8:0]	trigger_addr = 9'd0;
reg	[14:0]	post_trigger_counter = 15'd0;
always @(negedge sdram_clk) begin
	if(reset) begin
		halted				<=	1'b1;
		pre_trigger			<=	1'b0;
		armed					<=	1'b0;
		triggered			<=	1'b0;
	end
	else if(halted) begin
		if(run) begin
			halted			<=	1'b0;
			pre_trigger		<=	1'b1;
		end
	end
	else if(pre_trigger) begin
		if(rowaddr >= pre_trigger_count[14:0]) begin
			pre_trigger		<=	1'b0;
			armed				<=	1'b1;
		end
	end
	else if(armed) begin
		if(trigger) begin
			armed				<=	1'b0;
			triggered		<=	1'b1;
			trigger_page	<=	rowaddr;
			trigger_addr	<=	trigger_address_counter;
		end
	end
	else if(triggered) begin
		if(post_trigger_counter >= post_trigger_count[14:0]) begin
			triggered		<=	1'b0;
			halted			<=	1'b1;
		end
	end
end

//-----------------------------------
always @(negedge sdram_clk) begin
	reset <=  ep00wire[10];
end

//------------------------------------------------------------------------
// SDRAM transfer negotiator
//   This block handles communication between the SDRAM controller and
//   the FIFOs.  The FIFOs act as a simplified cache, holding at least
//   a full page on-chip while the PC reads the FIFO.  This dramatically
//   increases DRAM access performance since full pages can be read very
//   quickly.  Since the PC transfers are slower than the DRAM, there is
//   no fear of underrun.
//------------------------------------------------------------------------
parameter n_idle = 0,
          n_wackwait = 1,
          n_rackwait = 2,
			 n_busy = 3;
integer staten;
always @(negedge sdram_clk) begin
	cmd_pagewrite <= 1'b0;
	cmd_pageread <= 1'b0;
	if (reset == 1'b1) begin
		staten <= n_idle;
		rowaddr <= 15'd0;
		post_trigger_counter <= 15'd0;
	end
	else if (set_read_address) begin
		rowaddr <= read_address[14:0];
	end
	else begin
		case (staten)
			n_idle: begin
				staten <= n_idle;

				// If SDRAM WRITEs are enabled, trigger a block write whenever
				// the Pipe In buffer is at least 1/4 full (1 page, 512 words).
				if ((!halted) && (in_fifo_status[10:7] >= 4'b0100)) begin
					staten <= n_wackwait;
				end

				// If SDRAM READs are enabled, trigger a block read whenever
				// the Pipe Out buffer has room for at least 1 page (512 words).
				else if ((halted) && (epA0fifo_status[10:7] <= 4'b1000)) begin
					staten <= n_rackwait;
				end
			end


			n_wackwait: begin
				cmd_pagewrite <= 1'b1;
				staten <= n_wackwait;
				if (cmd_ack == 1'b1) begin
					rowaddr <= rowaddr + 1;
					if(triggered) begin
						post_trigger_counter	<=	post_trigger_counter + 1;
					end
					staten <= n_busy;
				end
			end
			

			n_rackwait: begin
				cmd_pageread <= 1'b1;
				staten <= n_rackwait;
				if (cmd_ack == 1'b1) begin
					rowaddr <= rowaddr + 1;
					staten <= n_busy;
				end
			end
			

			n_busy: begin
				staten <= n_busy;
				if (cmd_done == 1'b1) begin
					staten <= n_idle;
				end
			end

		endcase
	end
end


//------------------------------------------------------------------------
// SDRAM CONTROLLER
//------------------------------------------------------------------------
sdramctrl sdram (
		.clk(~sdram_clk),
		.clk_read(~sdram_clk),
		.reset(reset),
		.cmd_pagewrite(cmd_pagewrite),
		.cmd_pageread(cmd_pageread),
		.cmd_ack(cmd_ack),
		.cmd_done(cmd_done),
		.rowaddr_in(rowaddr),
		.fifo_din(in_fifo_dout),
		.fifo_read(in_fifo_read),
		.fifo_dout(out_fifo_din),
		.fifo_write(out_fifo_write),
		.sdram_cmd({sdram_cs_n, sdram_ras_n, sdram_cas_n, sdram_we_n}),
		.sdram_ba(sdram_ba),
		.sdram_a(sdram_a),
		.sdram_d(sdram_d));


fifo_2kx16 out_fifo (
	.rst				(read_address[15]),
	.wr_clk			(!sdram_clk),
	.rd_clk			(ti_clk),
	.din				(out_fifo_din),
	.wr_en			(out_fifo_write),
	.rd_en			(epA0_read),
	.dout				(epA0_din),
	.full				(),
	.empty			(),
	.rd_data_count	(),
	.wr_data_count	(epA0fifo_status)
);

//------------------------------------------------------------------------
// DCM
//   This ensures that the internal FPGA fabric clock (CLK) is phase 
//   aligned with the provided PLL clock (PLL_CLK) that is shared with 
//   the SDRAM at the board level.
//
//   The DCM here uses a phase delay to best align the fabric clock with
//   the SDRAM clock for optimum performance.  The phase delay was 
//   determined experimentally by testing the maximum "no-error" frequency
//   of RAMTester for various phase delays.
//------------------------------------------------------------------------
dcm_sys sdramDCM (
    .CLKIN_IN(clk1), 
    .RST_IN(1'b0), 
    .CLKIN_IBUFG_OUT(), 
    .CLK0_OUT(sdram_clk),
    .CLKFX_OUT(clk200) );

//------------------------------------------------------------------------
// USB HOST INTERFACE
//------------------------------------------------------------------------
okHostInterface okHI(
      .hi_in(hi_in), .hi_out(hi_out), .hi_inout(hi_inout),
      .ti_clk(ti_clk), .ok1(ok1), .ok2(ok2));

okWireIn ep00 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h00), .ep_dataout(ep00wire));
okWireIn ep01 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h01), .ep_dataout(pre_trigger_count));
okWireIn ep02 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h01), .ep_dataout(post_trigger_count));
okWireIn ep03 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h03), .ep_dataout(pattern_a));
okWireIn ep04 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h04), .ep_dataout(pattern_b));
okWireIn ep05 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h05), .ep_dataout(pattern_mask));
okWireIn ep06 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h06), .ep_dataout(read_address));
wire [15:0] ep20data = {triggered, armed, pre_trigger, halted, 3'b000, trigger_addr};
okWireOut ep20 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h20), .ep_datain(ep20data));
wire [15:0] ep21data = {1'b0, trigger_page};
okWireOut ep21 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h21), .ep_datain(ep21data));
okWireOut ep22 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h22), .ep_datain(pattern_a));
okWireOut ep23 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h23), .ep_datain(pattern_b));
okWireOut ep24 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h24), .ep_datain(pattern_mask));
okWireOut ep25 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h25), .ep_datain(comp_data));
okTriggerIn ep40 (.ok1(ok1), .ok2(ok2), .ep_addr(8'h40), .ep_clk(!sdram_clk), .ep_trigger(ep40_triggers));
okPipeOut epA0 (.ok1(ok1), .ok2(ok2), .ep_addr(8'ha0), .ep_read(epA0_read), .ep_datain(epA0_din));

endmodule
